//
// Autor: Rafael Alberto Moreno Parra
// Correo: ramsoftware@gmail.com
// Pgina Web: http://darwin.50webs.com
// Licencia: LGPL
// Fecha: Noviembre de 2012
//
unit Poblacion;

interface
uses
  Contnrs, Mersenne, Classes, Math, SysUtils, StrUtils;

const
  MAXIMAPOBLACION = 10000;  //Maxima poblacin que manejar el algoritmo gentico
  MAXPIEZASINDIV = 30;      //Mximo nmero de piezas que tendr cada individuo
  MAXNUMERODATOS = 1000;    //Mximo nmero de datos ledos para hacer regresin simblica
  MAXVARIABLEINDIV = 40;      //Mximo nmero de variables que tendr cada individuo
type
  //Registro de instrucciones (Variable que no tiene asignacin = operacin matemtica sencilla con variables que YA fueron asignadas)
  RInstrucciones = Record
    Funcion: integer;   //Se va a hacer uso de una funcin? 0 es no, 1 es seno, 2 es coseno, ....
    VarConA: boolean;  //Constante o variable en la primera parte?
    VarA: integer; //Cul variable de la primera parte
    ConA: real; //La constante de la primera parte
    Operador: integer; // +, -, *, /, ^
    VarConB: boolean;  //Constante o variable en la segunda parte?
    VarB: integer;  //Cul variable de la segunda parte
    ConB: real;  //La constante de la segunda parte
  end;

  //Qu es cada individuo
  RIndividuo = Record
    Generado: boolean; //TRUE el individuo ha sido generado
    Evaluado: boolean; //TRUE el individuo ya se le calcul el coeficiente de correlacin de Pearson
    Genotipo: Array[1..MAXPIEZASINDIV] of RInstrucciones;
    TotalPiezas: integer; //Total piezas(registros) que consta el individuo
    Aproximacion: double; //Almacena el coeficiente de correlacin de Pearson
  end;

  TPoblacion = class
  public
    //Individuos
    Individuos : array[1..MAXIMAPOBLACION] of RIndividuo;

    //Variables manejadas por los individuos
    VariablesIndiv : array[0..MAXVARIABLEINDIV] of double;
    ExpresionIndiv : array[0..MAXVARIABLEINDIV] of string;

    //Error matemtico que se puede dar en un individuo
    ERRORMATEMATICO: boolean;

    //Valores X,Y,Z que requiere la regresin simblica (lmite de 1000)
    valorX : array[1..MAXNUMERODATOS] of double;
    valorY : array[1..MAXNUMERODATOS] of double;
    valorZ : array[1..MAXNUMERODATOS] of double;
    limValores: integer;

    //Tamao de la poblacin
    tamPoblacion: integer;

    //Mejor individuo y mejor coeficiente
    mejorIndividuo: integer;
    mejorAproximacion: double;

    //Proceso de la regresin simblica
    Constructor Create();
    procedure ConfProceso(tamPoblacion: integer);
    procedure Proceso();
    function CalculaAproximacion(quien: integer): double;
    function Evaluar(quien: integer; valorX: double; valorY: double): double;
    procedure GeneraExpresion(quien: integer);
    function RetornaExpresion(quien: integer): string;

    //Mtodos para leer los datos del archivo
    procedure leeDatosTrios(URLArchivo: string);
    function TraerDatoCadena(Linea: string; Delimitador: char; NumeroToken: integer): double;

  end;

implementation

//Inicializa la simulacin
constructor TPoblacion.Create();
begin
  self.mejorIndividuo := -1;
  self.mejorAproximacion := 9999999999999;

  //Inicializa el generador de nmeros aleatorios Mersenne Twister
  Randomize;
  InitMT(Random(100000000));
end;


//Configura la regresin simblica
procedure TPoblacion.ConfProceso(tamPoblacion: integer);
var
  cont: integer;
begin
  //Configuracin de la poblacin
  self.tamPoblacion := tamPoblacion;

  //Inicializa a no evaluado, ni generado los individuos
  for cont := 1 to MAXIMAPOBLACION do
  begin
   Individuos[cont].Evaluado := false;
   Individuos[cont].Generado := false;
  end;

end;

{ Proceso principal
  El algoritmo es el siguiente:

  Se selecciona un individuo A al azar
  Se selecciona un individuo B al azar
  Se calcula la aproximacin de ambos individuos
  Quien tenga la mejor aproximacin (cercano a cero) es el mejor, se reproduce, sobreescribiendo las instrucciones
    del individuo peor. Este clon es mutado.
}
procedure TPoblacion.Proceso();
var
  indivA, indivB, mejorIndiv, peorIndiv, azarIndiv, pos, seleccionada: integer;
  aproxA, aproxB, mejorAprox: double;
begin
  //Seleccionar dos individuos al azar
  indivA := IRanMT mod tamPoblacion + 1;
  repeat
    indivB := IRanMT mod tamPoblacion + 1;
  until indivA <> indivB;

  //Al iniciar la simulacin, los individuos no estn formados an, as que
  //al ser seleccionados se crean
  if Individuos[indivA].Generado = false then GeneraExpresion(indivA);

  //Si el individuo an no se le calcula la aproximacin se procede a hacerlo
  //Se queda en un ciclo continuo de generacin hasta que el individuo no genere errores matemticos
  if Individuos[indivA].Evaluado = false then
    repeat
        Individuos[indivA].Aproximacion := CalculaAproximacion(indivA);
        if ERRORMATEMATICO = true then GeneraExpresion(indivA);
    until Individuos[indivA].Evaluado = true;
  aproxA := Individuos[indivA].Aproximacion;

  if Individuos[indivB].Generado = false then GeneraExpresion(indivB);
  if Individuos[indivB].Evaluado = false then
    repeat
        Individuos[indivB].Aproximacion := CalculaAproximacion(indivB);
        if ERRORMATEMATICO = true then GeneraExpresion(indivB);
    until Individuos[indivB].Evaluado = true;
  aproxB := Individuos[indivB].Aproximacion;

  //Deduce el mejor y peor individuo
  if (aproxA < aproxB) or ((aproxA = aproxB) and (self.mejorIndividuo = indivA)) then
  begin
      mejorIndiv := indivA;
      mejorAprox := aproxA;
      peorIndiv := indivB;
  end
  else
  begin
      mejorIndiv := indivB;
      mejorAprox := aproxB;
      peorIndiv := indivA;
  end;

  //Detecta si es el mejor individuo generado en la simulacin
  if mejorAprox < self.mejorAproximacion then
  begin
    self.mejorAproximacion := mejorAprox;
    self.mejorIndividuo := mejorIndiv;
  end;

  //El mejor individuo se reproduce sobreescribiendo el peor
  for pos := 1 to Individuos[mejorIndiv].TotalPiezas do
  begin
      Individuos[peorIndiv].Genotipo[pos].Funcion := Individuos[mejorIndiv].Genotipo[pos].Funcion;
      Individuos[peorIndiv].Genotipo[pos].VarConA := Individuos[mejorIndiv].Genotipo[pos].VarConA;
      Individuos[peorIndiv].Genotipo[pos].VarA := Individuos[mejorIndiv].Genotipo[pos].VarA;
      Individuos[peorIndiv].Genotipo[pos].ConA := Individuos[mejorIndiv].Genotipo[pos].ConA;
      Individuos[peorIndiv].Genotipo[pos].Operador := Individuos[mejorIndiv].Genotipo[pos].Operador;
      Individuos[peorIndiv].Genotipo[pos].VarConB := Individuos[mejorIndiv].Genotipo[pos].VarConB;
      Individuos[peorIndiv].Genotipo[pos].VarB := Individuos[mejorIndiv].Genotipo[pos].VarB;
      Individuos[peorIndiv].Genotipo[pos].ConB := Individuos[mejorIndiv].Genotipo[pos].ConB;
  end;
  Individuos[peorIndiv].TotalPiezas := Individuos[mejorIndiv].TotalPiezas;

  //Se muta el hijo (clon del mejor)
  seleccionada := IRanMT mod Individuos[peorIndiv].TotalPiezas + 1;
  Individuos[peorIndiv].Genotipo[seleccionada].Funcion := IRanMT mod 8;
  Individuos[peorIndiv].Genotipo[seleccionada].VarA := IRanMT mod seleccionada;
  Individuos[peorIndiv].Genotipo[seleccionada].ConA := IRanMT / IRanMT;
  Individuos[peorIndiv].Genotipo[seleccionada].Operador := IRanMT mod 4;
  Individuos[peorIndiv].Genotipo[seleccionada].VarB := IRanMT mod seleccionada;
  Individuos[peorIndiv].Genotipo[seleccionada].ConB := IRanMT / IRanMT;
  Individuos[peorIndiv].Evaluado := false;

end;

//Genera al azar el individuo
procedure TPoblacion.GeneraExpresion(quien: integer);
var
  pos: integer;
begin
  Individuos[quien].Evaluado := false;
  Individuos[quien].Generado := true;
  Individuos[quien].totalPiezas := IRanMT mod MAXPIEZASINDIV + 1;

  //Construye instruccin por instruccin
  for pos := 2 to Individuos[quien].totalPiezas do
  begin
      //Posible funcin
      if IRanMT mod 100 < 50 then
        Individuos[quien].Genotipo[pos].Funcion := IRanMT mod 8
      else
        Individuos[quien].Genotipo[pos].Funcion := 0;

      //Define si la primera parte es variable o constante
      if IRanMT mod 100 < 50 then
        Individuos[quien].Genotipo[pos].VarConA := true
      else
        Individuos[quien].Genotipo[pos].VarConA := false;

      //La variable de la primera parte
      Individuos[quien].Genotipo[pos].VarA := IRanMT mod pos;

      //Valor real de la primera parte
      Individuos[quien].Genotipo[pos].ConA := IRanMT / IRanMT;

      //Operador
      Individuos[quien].Genotipo[pos].Operador := IRanMT mod 4;  //+  -  * / 

      //Define si la segunda parte es variable o constante
      if IRanMT mod 100 < 50 then
        Individuos[quien].Genotipo[pos].VarConB := true
      else
        Individuos[quien].Genotipo[pos].VarConB := false;

      //La variable de la segunda parte
      Individuos[quien].Genotipo[pos].VarB := IRanMT mod pos;

      //Valor real de la segunda parte
      Individuos[quien].Genotipo[pos].ConB := IRanMT / IRanMT;
  end;
end;

//Retorna la aproximacin de un determinado individuo
function TPoblacion.CalculaAproximacion(quien: integer): double;
var
  pos: integer;
  valor, AcumulaIndividuo: double;
begin
  Individuos[quien].Evaluado := false;
  try

    AcumulaIndividuo := 0;
    for pos := 1 to self.limValores do
    begin
			valor := Evaluar(quien, valorX[pos], valorY[pos]);
      if ERRORMATEMATICO then
      begin
        Result := -1;
        Exit;
      end;
      //Ajuste de mnimos cuadrados (tomado de: http://iqc.udg.es/~perico/docencia/QTC/ajustes.pdf )
      AcumulaIndividuo := AcumulaIndividuo + (valorZ[pos] - valor) * (valorZ[pos] - valor);
    end;

	except

		on E : Exception do
    begin
      ERRORMATEMATICO := true;
      Result := -1;
      Exit;
    end;
  end;
  Individuos[quien].Evaluado := true;
  CalculaAproximacion := AcumulaIndividuo;
end;

//Evala el individuo yendo de instruccin en instruccin que lo compone
function TPoblacion.Evaluar(quien: integer; valorX: double; valorY: double): double;
var
  pos: integer;
  parte1, parte2, acumula: double;
begin
  VariablesIndiv[0] := valorX;
  VariablesIndiv[1] := valorY;
  acumula := 0;

  ERRORMATEMATICO := false;
  try
    for pos := 2 to Individuos[quien].TotalPiezas do
    begin
      if Individuos[quien].Genotipo[pos].VarConA = true then
        parte1 := VariablesIndiv[Individuos[quien].Genotipo[pos].VarA]
      else
        parte1 := Individuos[quien].Genotipo[pos].ConA;

      if Individuos[quien].Genotipo[pos].VarConB = true then
        parte2 := VariablesIndiv[Individuos[quien].Genotipo[pos].VarB]
      else
        parte2 := Individuos[quien].Genotipo[pos].ConB;

      //Operacin bsica
      case Individuos[quien].Genotipo[pos].Operador of
        0: begin acumula := parte1 + parte2; end;
        1: begin acumula := parte1 - parte2; end;
        2: begin acumula := parte1 * parte2; end;
        3: begin
            if parte2 <> 0 then
              acumula := parte1 / parte2
            else
            begin
              ERRORMATEMATICO := true;
              Evaluar := 0;
              Exit
            end;
          end;
      end;

      //Funciones
      case Individuos[quien].Genotipo[pos].Funcion of
        0: begin VariablesIndiv[pos] := acumula; end;
        1: begin VariablesIndiv[pos] := sin(acumula); end;
        2: begin VariablesIndiv[pos] := cos(acumula); end;
        3: begin VariablesIndiv[pos] := tan(acumula); end;
        4: begin VariablesIndiv[pos] := abs(acumula); end;
        5: begin VariablesIndiv[pos] := arctan(acumula); end;
        6:  begin
              if acumula <> 0 then
                VariablesIndiv[pos] := ln(abs(acumula))
              else
              begin
                ERRORMATEMATICO := true;
                Evaluar := 0;
                Exit;
              end;
            end;
        7: begin VariablesIndiv[pos] := sqrt(abs(acumula)); end;
      end;

    end;
    //Resultado final
    Evaluar := VariablesIndiv[Individuos[quien].TotalPiezas];
  except
    on E : Exception do
    begin
      ERRORMATEMATICO := true;
      Evaluar := 0;
      Exit;
    end;
  end;
end;


//Retorna la expresin de ese individuo usando el mismo mtodo de evaluacin
//pero en vez de acumular valores, acumula cadenas
function TPoblacion.RetornaExpresion(quien: integer): string;
var
  pos, varA, varB: integer;
  parte1, parte2, acumula: string;
begin
    ExpresionIndiv[0] := 'x';
	ExpresionIndiv[1] := 'y';

    for pos := 2 to Individuos[quien].TotalPiezas do
    begin
      if Individuos[quien].Genotipo[pos].VarConA = true then
      begin
        varA := Individuos[quien].Genotipo[pos].VarA;
        parte1 := ExpresionIndiv[varA];
      end
      else
        parte1 := floattostr(Individuos[quien].Genotipo[pos].ConA);

      if Individuos[quien].Genotipo[pos].VarConB = true then
      begin
        varB := Individuos[quien].Genotipo[pos].VarB;
        parte2 := ExpresionIndiv[varB];
      end
      else
        parte2 := floattostr(Individuos[quien].Genotipo[pos].ConB);

      case Individuos[quien].Genotipo[pos].Operador of
        0: begin acumula := parte1 + '+' + parte2; end;
        1: begin acumula := parte1 + '-' + parte2; end;
        2: begin acumula := parte1 + '*' + parte2; end;
        3: begin acumula := parte1 + '/' + parte2; end;
      end;

      case Individuos[quien].Genotipo[pos].Funcion of
        0: begin acumula := '(' + acumula + ')'; end;
        1: begin acumula := 'seno(' + acumula + ')'; end;
        2: begin acumula := 'cos(' + acumula + ')'; end;
        3: begin acumula := 'tan(' + acumula + ')'; end;
        4: begin acumula := 'abs(' + acumula + ')'; end;
        5: begin acumula := 'atan(' + acumula+ ')'; end;
        6: begin acumula := 'ln(abs(' + acumula + '))'; end;
        7: begin acumula := 'raiz(abs(' + acumula + '))'; end;
      end;
      
      //Expresin completa
      ExpresionIndiv[pos] := acumula;
    end;
  RetornaExpresion := ExpresionIndiv[Individuos[quien].TotalPiezas];
end;


//Lee los datos de pares de valores de un archivo plano
procedure TPoblacion.leeDatosTrios(URLArchivo: string);
var
  archivo: TextFile;
  leelinea: string;
  numX, numY, numZ: double;
begin
    //Prepara para lectura
    AssignFile(archivo, URLArchivo);
    Reset(archivo);

    //Lee lnea a lnea el archivo
    limValores := 0;
    while not Eof(archivo) do
    begin
      ReadLn(archivo, leelinea);

      //Interpreta la lnea para extraer los valores X, Y
      numX := TraerDatoCadena(leelinea, ';', 1);
      numY := TraerDatoCadena(leelinea, ';', 2);
      numZ := TraerDatoCadena(leelinea, ';', 3);

      //Lo agrega al listado de pares valores
      Inc(limValores);
      self.valorX[limValores] := numX;
      self.valorY[limValores] := numY;
			self.valorZ[limValores] := numZ;
    end;
    Close(archivo);
end;

//Dada una cadena con separaciones por delimitador, trae determinado tem
function TPoblacion.TraerDatoCadena(Linea: string; Delimitador: char; NumeroToken: integer): double;
var
  cont, numTrae: integer;
  numero: string;
begin
  numero := '';
  numTrae := 0;
  for cont := 1 to length(Linea) do
  begin
    if Linea[cont]<>Delimitador then
      numero := numero + Linea[cont]
    else
    begin
      numTrae := numTrae + 1;
      if numTrae = NumeroToken then
      begin
        numero := trim(numero);
        Result := strtofloat(numero);
        exit;
      end;
      numero := '';
    end;
  end;
  numero := trim(numero);
  Result := strtofloat(numero);
end;

end.
